
/*
Exemplo cliente-servidor
Codigo do cliente
(C) 2011-2012 Pedro Freire

Compilar (em Mac/Unix) com
	gcc -Wall client.c -o client
ou em Windows com
	gcc -Wall client.c -o client -lws2_32

Correr com
	client <fqdn-do-servidor>:<porta-do-servidor>
ou
	client <fqdn-do-servidor>
ou
	client
(esta ultima liga-se a um servidor a correr no mesmo computador e
que foi arrancado sem argumentos, ou seja, com a porta de omissao)
Vai varrer PORT_SCAN_NUM portas a partir da indicada 'a procura de
servidores ligados, e liga-se a todos os que encontrar.
*/


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include <assert.h>

#ifdef WIN32
#include <windows.h>
#include <winsock2.h>
#ifdef errno
#undef errno
#endif
#define EADDRNOTAVAIL		WSAEADDRNOTAVAIL
#define EADDRINUSE		WSAEADDRINUSE
#define ETIMEDOUT		WSAETIMEDOUT
#define ECONNREFUSED		WSAECONNREFUSED
#define ECONNRESET		WSAECONNRESET
#define perror_and_exit(msg)	{ fprintf(stderr, "Erro %i do %s do Windows Sockets. Visita\nhttp://msdn.microsoft.com/en-us/library/windows/desktop/ms740668(v=vs.85).aspx\npara saber a descricao.\n", (int) WSAGetLastError(), msg); exit(EXIT_FAILURE); }
#define errno			WSAGetLastError()
#define sleep(s)		Sleep((s)*1000)
typedef int socklen_t;
#define IPPORT_USERRESERVED	5000
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#define INVALID_SOCKET		-1
#define SOCKET_ERROR		-1
#define perror_and_exit(msg)	{ perror(msg); exit(EXIT_FAILURE); }
#endif


#define DEFAULT_HOST	"localhost"
#define DEFAULT_PORT	IPPORT_USERRESERVED

/* Numero de portas rastreadas 'a procura de servidores disponiveis
   (em Windows este processo e' muito lento pelo que rastreamos menos;
   isto tambem significa que em Windows suportamos menos servidores)
*/
#ifdef WIN32
#define PORT_SCAN_NUM	  5
#else
#define PORT_SCAN_NUM	100
#endif


int main( int argc, char *argv[] )
{
	char                *hostname;
	uint16_t            port;
	char                buffer[1024+1];
	char                *p;
	int                 server_count;
	int                 server_socket;
	int                 server_sockets[ PORT_SCAN_NUM ];
	struct sockaddr_in  server_sa, servers_sa;
	struct hostent      *hp;
	int                 recv_sz;
	int                 num;
	fd_set              server_set, read_set;

#ifdef WIN32
	WSADATA wsaData;
	int wsaErr = WSAStartup( MAKEWORD(2,2), &wsaData );
	if( wsaErr != 0 )
		{
		fprintf( stderr, "Erro %i do WSAStartup() do Windows Sockets. Visita\nhttp://msdn.microsoft.com/en-us/library/windows/desktop/ms742213(v=vs.85).aspx\npara saber a descricao.\n", (int) wsaErr );
		exit( EXIT_FAILURE );
		}
#else
	signal( SIGPIPE, SIG_IGN );
#endif

	/* Establecer o hostname (FQDN) e a porta IP para o servidor
	*/
	hostname = (argc > 1 ? argv[1] : DEFAULT_HOST );
	port = DEFAULT_PORT;
	p = strchr( hostname, ':' );
	if( p != NULL )
		{
		port = (uint16_t) atol( p+1 );
		*p = '\0';
		}
	if( port < IPPORT_USERRESERVED )
		{
		fprintf( stderr, "Porta incorrecta (%u). Deveria ser >= a %u para evitar conflitos.\n", port, IPPORT_USERRESERVED );
		exit( EXIT_FAILURE );
		}

	/* Determinar o IPv4 do hostname (FQDN)
	*/
	hp = gethostbyname( hostname );
	if( hp == NULL )
		perror_and_exit( "gethostbyname()" );
	if( hp->h_addrtype == AF_INET )
		{
		memset( &servers_sa, 0, sizeof(servers_sa) );
		servers_sa.sin_family = AF_INET;
		memcpy( &servers_sa.sin_addr, hp->h_addr, hp->h_length );
		servers_sa.sin_port = htons( port );
		}
	else
		{
		fprintf( stderr, "O servidor '%s' tem um endereco de uma familia desconhecida.\n", hostname );
		exit( EXIT_FAILURE );
		}

	printf( "** Vou rastrear a existencia de servidores em %d portas...\n", PORT_SCAN_NUM );
	FD_ZERO( &server_set );  /* apaga (torna vazio) o conjunto "server_set" */
	server_count = 0;
	for( num = PORT_SCAN_NUM;  num > 0;  num-- )
		{
		/* Obter um socket (uma "tomada") para falar IPv4
		*/
		server_socket = socket( PF_INET, SOCK_STREAM, 0 );
		if( server_socket == INVALID_SOCKET )
			perror_and_exit( "socket()" );

		/* Estabelecer a ligacao
		*/
		server_sa = servers_sa;
		server_sa.sin_port = htons( port++ );
		if( connect(server_socket, (struct sockaddr*) &server_sa, sizeof(server_sa)) == SOCKET_ERROR )
			{
			if( (errno != EADDRNOTAVAIL  &&
			     errno != ETIMEDOUT      &&
			     errno != ECONNREFUSED)      ||
			    port == (uint16_t) 0x0000 )
				perror_and_exit( "connect()" );

			/* Fechar e re-establecer novo socket, senao o connect()
			   da proxima vez da'-me "invalid argument"
			*/
			shutdown( server_socket, 2 );
			continue;
			}

		/* Consegui ligacao!
		*/
		FD_SET( server_socket, &server_set );  /* guarda este socket no conjunto "server_set" */
		server_sockets[ server_count++ ] = server_socket;

		/* Informar o utilizador de onde esta' este servidor ligado
		*/
		if( server_sa.sin_family == AF_INET )
			{
			printf( "** Liguei-me com o socket %d ao servidor '%s' no endereco IPv4 %s:%u...\n",
				server_socket, hostname,
				inet_ntoa(server_sa.sin_addr), ntohs(server_sa.sin_port) );
			}
		else
			printf( "** Liguei-me com o socket %d ao servidor '%s' numa familia de enderecos que nao IPv4...\n",
				server_socket, hostname );
		}

	printf( "** Terminei o rastreio, estou a ouvir o(s) servidor(es)...\n" );
	while( server_count > 0 )
		{
		/* Espera ate' que um dos servidores tenha uma mensagem a enviar
		*/
		read_set = server_set;
		if( select(FD_SETSIZE, &read_set, NULL, NULL, NULL) == SOCKET_ERROR )
			perror_and_exit( "select()" );

		for( num = 0; num < server_count; num++ )
			{
			server_socket = server_sockets[ num ];
			if( !FD_ISSET(server_socket, &read_set) )
				continue;

			/* Ler mensagem do servidor
			*/
			recv_sz = recv( server_socket, buffer, sizeof(buffer), 0 );
			if( recv_sz == SOCKET_ERROR )
				{
				if( errno != EPIPE  &&  errno != ECONNRESET )
					perror_and_exit( "recv()" );  /* houve um erro, ou... */
				/* ...o outro lado desligou-se */
				recv_sz = 0;
				}
			if( recv_sz == 0 )
				{
				printf( "\n** O servidor no socket %d desligou a ligacao.\n", server_socket );
				shutdown( server_socket, 2 );
				FD_CLR( server_socket, &server_set );
				memmove( &server_sockets[num], &server_sockets[num+1], (server_count-1-num)*sizeof(server_sockets[0]) );
				num--;  /* para contrabalancar o num++ ao fim do for() */
				server_count--;
				}
			else
				{
				buffer[recv_sz] = '\0';  /* assegurar-me que buffer e' uma string em C */
				printf( "** Servidor com socket %d diz: %s\n", server_socket, buffer );
				}
			}
		}

	/* Terminar
	*/
	puts( "\n** Todos os servidores desligaram." );
#ifdef WIN32
	WSACleanup();
#endif
	return EXIT_SUCCESS;
}
