
/*
Exemplo cliente-servidor
Codigo do servidor
(C) 2011-2012 Pedro Freire

Compilar (em Mac/Unix) com
	gcc -Wall server.c -o server
ou em Windows com
	gcc -Wall server.c -o server -lws2_32

Correr com
	server <porta-do-servidor>
ou
	server
(esta ultima cria um servidor com um numero de porta automatico)
Se a porta em questao nao estiver disponivel, tenta a seguinte
e assim por diante, ate' um maximo de PORT_SCAN_NUM tentativas.

E' possivel falar com este servidor usando uma ligacao Telnet normal.

Atencao que e' importante permitir que a firewall do sistema operativo
crie uma excepcao para o servidor, senao ele nao vai receber ligacoes!
*/


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <errno.h>
#include <assert.h>

#ifdef WIN32
#include <windows.h>
#include <winsock2.h>
#ifdef errno
#undef errno
#endif
#define EADDRNOTAVAIL		WSAEADDRNOTAVAIL
#define EADDRINUSE		WSAEADDRINUSE
#define ETIMEDOUT		WSAETIMEDOUT
#define ECONNREFUSED		WSAECONNREFUSED
#define ECONNRESET		WSAECONNRESET
#define perror_and_exit(msg)	{ fprintf(stderr, "Erro %i do %s do Windows Sockets. Visita\nhttp://msdn.microsoft.com/en-us/library/windows/desktop/ms740668(v=vs.85).aspx\npara saber a descricao.\n", (int) WSAGetLastError(), msg); exit(EXIT_FAILURE); }
#define errno			WSAGetLastError()
#define sleep(s)		Sleep((s)*1000)
typedef int socklen_t;
#define IPPORT_USERRESERVED	5000
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#define INVALID_SOCKET		-1
#define SOCKET_ERROR		-1
#define perror_and_exit(msg)	{ perror(msg); exit(EXIT_FAILURE); }
#endif


#define DEFAULT_PORT	IPPORT_USERRESERVED
#define PORT_SCAN_NUM	100  /* numero de portas rastreadas 'a procura de uma disponivel */

#define TEXT_WELCOME	"Ligado ao servidor!\n"

#define TEXT_HELP	"\n"							\
			"Nao vou ler o que o cliente escreve.\n"				\
			"Vou em vez disso, a cada 1s-5s dizer 'ping'.\n"		\
			"\n"


int main( int argc, char *argv[] )
{
	uint16_t            port;
	double              num;
	int                 server_socket, client_socket;
	struct sockaddr_in  server_sa;  /* sockaddr deste servidor */
	struct sockaddr_in  client_sa;  /* sockaddr do cliente */
	socklen_t           client_sa_sz;

#ifdef WIN32
	WSADATA wsaData;
	int wsaErr = WSAStartup( MAKEWORD(2,2), &wsaData );
	if( wsaErr != 0 )
		{
		fprintf( stderr, "Erro %i do WSAStartup() do Windows Sockets. Visita\nhttp://msdn.microsoft.com/en-us/library/windows/desktop/ms742213(v=vs.85).aspx\npara saber a descricao.\n", (int) wsaErr );
		exit( EXIT_FAILURE );
		}
#else
	signal( SIGPIPE, SIG_IGN );
#endif

	/* Establecer a porta IP para o servidor
	*/
	port = (argc > 1 ? (uint16_t) atol(argv[1]) : DEFAULT_PORT );
	if( port < IPPORT_USERRESERVED )
		{
		fprintf( stderr, "Porta incorrecta (%u). Deveria ser >= a %u para evitar conflitos.\n", port, IPPORT_USERRESERVED );
		exit( EXIT_FAILURE );
		}

	/* Obter um socket (uma "tomada") para falar IPv4
	*/
	server_socket = socket( PF_INET, SOCK_STREAM, 0 );
	if( server_socket == INVALID_SOCKET )
		perror_and_exit( "socket()" );

	/* Atribuir (bind) o endereco IPv4 deste servidor a este socket;
	   Com isto o socket passa a estar acessivel externamente.
	   Se a porta em questao nao estiver disponivel, tenta a seguinte
	   e assim por diante, ate' um maximo de PORT_SCAN_NUM tentativas.
	*/
	for( num = PORT_SCAN_NUM;  num > 0;  num-- )
		{
		memset( &server_sa, 0, sizeof(server_sa) );
		server_sa.sin_family = AF_INET;
		server_sa.sin_addr.s_addr = htonl( INADDR_ANY );
		server_sa.sin_port = htons( port );
		if( bind(server_socket, (struct sockaddr*) &server_sa, sizeof(server_sa)) != SOCKET_ERROR )
			break;
		if( errno != EADDRINUSE  ||  port == (uint16_t) 0xFFFF )
			perror_and_exit( "bind()" );
		port++;
		}

	/* Permitir que os clientes se liguem (1 de cada vez) a este socket;
	   Atencao que isto so' funcionara' se for aberta esta porta na
	   firewall do sistema!
	*/
	if( listen(server_socket, 1) == SOCKET_ERROR )
		perror_and_exit( "listen()" );

	/* Informar o utilizador de onde esta' este servidor ligado
	*/
	if( server_sa.sin_family == AF_INET )
		{
		printf( "Estou 'a espera de ligacoes na porta IPv4 %u...\n\n",
			ntohs(server_sa.sin_port) );
			/* o IP e' exibido como "0.0.0.0" */
		}
	else
		puts( "Estou 'a espera de ligacoes numa familia de enderecos que nao IPv4...\n" );


	for(;;)
		{
		/* Esperar que um cliente se ligue
		*/
		client_sa_sz = sizeof( client_sa );
		client_socket = accept( server_socket, (struct sockaddr*) &client_sa, &client_sa_sz );
		if( client_socket == INVALID_SOCKET )
			perror_and_exit( "accept()" );

		/* Exibir informacao sobre quem se ligou
		*/
		if( client_sa.sin_family == AF_INET )
			{
			printf( "Ligou-se um cliente a partir do endereco IPv4 %s:%u\n",
				inet_ntoa(client_sa.sin_addr), ntohs(client_sa.sin_port) );
			}
		else
			puts( "Ligou-se um cliente a partir de uma familia de enderecos que nao IPv4" );

		/* Envia mensagem de boas-vindas e ajuda
		   e se nao houver erro, executa o ciclo principal
		*/
		srand( (unsigned int) time(NULL) + (port & 0x1F) * 3 );
		if( send(client_socket, TEXT_WELCOME TEXT_HELP, strlen(TEXT_WELCOME TEXT_HELP), 0) == INVALID_SOCKET )
			{
			if( errno != EPIPE  &&  errno != ECONNRESET )
				perror_and_exit( "send()" );  /* houve um erro, ou... */
			/* ...o outro lado desligou-se */
			}
		else for(;;)  /* forever */
			{
			/* Faz uma pausa de 1 a 5 segundos
			*/
			sleep( 1 + (rand() % 5) );

			/* Envia o 'ping'
			*/
			fputs( "A enviar 'ping'... ", stdout );
			if( send(client_socket, "ping", strlen("ping"), 0) == INVALID_SOCKET )
				{
				if( errno != EPIPE  &&  errno != ECONNRESET )
					perror_and_exit( "send()" );  /* houve um erro, ou... */
				/* ...o outro lado desligou-se */
				break;
				}
			puts( "Enviado." );
			}

		/* Exibir informacao sobre quem se desligou
		*/
		if( client_sa.sin_family == AF_INET )
			{
			printf( "\nDesligou-se o cliente no endereco IPv4 %s:%u\n\n",
				inet_ntoa(client_sa.sin_addr), ntohs(client_sa.sin_port) );
			}
		else
			puts( "\nDesligou-se o cliente de uma familia de enderecos que nao IPv4\n" );

        	/* Fechar o socket do cliente
        	*/
		shutdown( client_socket, 2 );
		}

	/* Nunca chega aqui; mas seria assim:
	*/
       	shutdown( server_socket, 2 );
#ifdef WIN32
	WSACleanup();
#endif
	return EXIT_SUCCESS;
}
